const { assign: _assign } = Object;
const { declare } = require('@babel/helper-plugin-utils');
// Supported plugins
const pluginModuleResolver = require('babel-plugin-module-resolver');
const pluginImport = require('babel-plugin-import');
const pluginEmotion = require('babel-plugin-emotion');
const pluginStyledComponents = require('babel-plugin-styled-components');
// Supported persets
const presetEnv = require('@babel/preset-env');
const presetReact = require('@babel/preset-react');
const presetTypeScript = require('@babel/preset-typescript');
const presetMinify = require('babel-preset-minify');
// Supported TC39 proposals
const dynamicImport = require('@babel/plugin-syntax-dynamic-import');
const importMeta = require('@babel/plugin-syntax-import-meta');
const classProperties = require('@babel/plugin-proposal-class-properties');
const numericSeparator = require('@babel/plugin-proposal-numeric-separator');
const privateMethods = require('@babel/plugin-proposal-private-methods');
const decorators = require('@babel/plugin-proposal-decorators');
const nullishCoalescingOperator = require('@babel/plugin-proposal-nullish-coalescing-operator');
const optionalChaining = require('@babel/plugin-proposal-optional-chaining');
const throwExpressions = require('@babel/plugin-proposal-throw-expressions');
const doExpressions = require('@babel/plugin-proposal-do-expressions');
const exportDefaultFrom = require('@babel/plugin-proposal-export-default-from');
const exportNamespaceFrom = require('@babel/plugin-proposal-export-namespace-from');
const logicalAssignmentOperators = require('@babel/plugin-proposal-logical-assignment-operators');
const pipelineOperator = require('@babel/plugin-proposal-pipeline-operator');
const partialApplication = require('@babel/plugin-proposal-partial-application');
// React optimized plugins
const reactRemovePropTypes = require('babel-plugin-transform-react-remove-prop-types');
const reactConstantElements = require('@babel/plugin-transform-react-constant-elements');
const reactInlineElements = require('@babel/plugin-transform-react-inline-elements');

// Default legacy (not modern) targets
const legacyTargets = {
  chrome: 49,
  firefox: 52,
  ie: 11,
};

// Supported extensions
const extensions = ['.js', '.jsx', '.ts', '.tsx', '.mjs'];

module.exports = declare((api, options) => {
  // Declare required version
  api.assertVersion('^7.5.0');

  // Cache in development mode
  const isDev = api.env('development');
  api.cache.using(() => isDev);

  // Get user options with default
  const {
    modern = true,
    react = true,
    typescript = true,
    minify = false,
    alias = false,
    styledComponents = false,
    emotion = false,
    imported = {
      antd: false,
      lodash: false,
      materialUi: false,
    },
    proposals = {
      'dynamic-import': true, // ················ Stage 4
      'import-meta': true, // ··················· Stage 3
      'class-properties': true, // ·············· Stage 3
      'numeric-separator': true, // ············· Stage 3
      'private-methods': true, // ··············· Stage 3
      decorators: true, // ······················ Stage 2
      'nullish-coalescing-operator': true, // ··· Stage 2
      'optional-chaining': true, // ············· Stage 2
      'throw-expressions': true, // ············· Stage 2
      'do-expressions': true, // ················ Stage 1
      'export-default-from': true, // ··········· Stage 1
      'export-namespace-from': true, // ········· Stage 1
      'logical-assignment-operators': true, // ·· Stage 1
      'pipeline-operator': true, // ············· Stage 1
      'partial-application': true, // ··········· Stage 1
    },
    ...presetEnvOptions
  } = options;

  // Initial presets array with preset-env
  const presets = [
    [
      presetEnv.default,
      _assign(
        {
          loose: true,
          modules: false,
          useBuiltIns: false,
          targets: modern ? { esmodules: modern } : legacyTargets,
        },
        presetEnvOptions,
      ),
    ],
  ];

  // Initial babel plugins (macros supported)
  let plugins = [];

  // Add module resolver plugin for alias
  if (alias) {
    const moduleResolverOptions = {
      root: ['.'],
      alias: { '@/': './src/' },
      extensions,
      stripExtensions: extensions,
    };
    plugins.push([pluginModuleResolver.default, _assign(moduleResolverOptions, alias)]);
  }

  // Add import optimization plugin with antd
  if (imported.antd) {
    const antdOptions = {
      libraryName: 'antd',
      libraryDirectory: 'lib',
      style: true,
    };
    plugins.push([pluginImport.default, _assign(antdOptions, imported.antd), 'antd']);
  }

  // Add import optimization plugin with lodash
  if (imported.lodash) {
    const lodashOptions = {
      libraryName: 'lodash',
      libraryDirectory: '',
      camel2DashComponentName: false,
    };
    plugins.push([pluginImport.default, _assign(lodashOptions, imported.lodash), 'lodash']);
  }

  // Add import optimization plugin with material-ui
  if (imported.materialUi) {
    const materialUiOptions = {
      libraryName: '@material-ui/core',
      libraryDirectory: 'components',
      camel2DashComponentName: false,
    };
    plugins.push([
      pluginImport.default,
      _assign(materialUiOptions, imported.materialUi),
      'material-ui',
    ]);
  }

  // Add Emotion plugin
  if (emotion) {
    const emotionOptions = {
      sourceMap: isDev,
      autoLabel: isDev,
      labelFormat: '[filename]-[local]',
      cssPropOptimization: true,
    };
    plugins.push([pluginEmotion, _assign(emotionOptions, emotion)]);
  }

  // Add styled-components plugin
  if (styledComponents) {
    const styledComponentsOptions = {
      displayName: isDev,
      fileName: isDev,
      minify: !isDev,
      transpileTemplateLiterals: !isDev,
      pure: !isDev,
    };
    plugins.push([pluginStyledComponents, _assign(styledComponentsOptions, styledComponents)]);
  }

  // Add react preset
  if (react) {
    const reactOptions = { development: isDev };
    presets.push([presetReact.default, _assign(reactOptions, react)]);
  }

  // Add typescript preset
  if (typescript) {
    const typescriptOptions = { isTSX: !!react, allExtensions: !!react };
    presets.push([presetTypeScript.default, _assign(typescriptOptions, typescript)]);
  }

  // Enabled TC39 proposals with options
  const enabledProposals = [
    ['dynamic-import', [dynamicImport, {}]],
    ['import-meta', [importMeta, {}]],
    // 'decorators' plugin must come before 'class-properties' plugin
    ['decorators', [decorators, { decoratorsBeforeExport: true }]],
    ['class-properties', [classProperties, { loose: true }]],
    ['numeric-separator', [numericSeparator, {}]],
    ['private-methods', [privateMethods, { loose: true }]],
    ['nullish-coalescing-operator', [nullishCoalescingOperator, { loose: true }]],
    ['optional-chaining', [optionalChaining, { loose: true }]],
    ['throw-expressions', [throwExpressions, {}]],
    ['do-expressions', [doExpressions, {}]],
    ['export-default-from', [exportDefaultFrom, {}]],
    ['export-namespace-from', [exportNamespaceFrom, {}]],
    ['logical-assignment-operators', [logicalAssignmentOperators, {}]],
    ['pipeline-operator', [pipelineOperator, { proposal: 'minimal' }]],
    ['partial-application', [partialApplication, {}]],
  ]
    .filter(proposal => proposals[proposal[0]])
    .map(proposal => [proposal[1][0].default, _assign(proposal[1][1], proposals[proposal[0]])]);

  plugins = [...plugins, ...enabledProposals];

  if (!isDev) {
    // Optimize react in production
    if (react) {
      plugins.push([reactRemovePropTypes.default, { ignoreFilenames: ['node_modules'] }]);
      plugins.push([reactConstantElements.default, { allowMutablePropsOnTags: [] }]);
      plugins.push([reactInlineElements.default, {}]);
    }
    // Add minify preset in production
    if (minify) {
      const minifyOptions = { removeConsole: true, removeDebugger: true };
      presets.push([presetMinify.default, _assign(minifyOptions, minify)]);
    }
  }

  return {
    presets,
    plugins,
    sourceMaps: isDev ? 'inline' : false,
    compact: !isDev,
    minified: !isDev,
    comments: isDev,
  };
});
